defaultSymbols = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"

def get_colorDict(nestedColorsList, symbols):
    colorDict = {}
    i = 0
    for colors in nestedColorsList:
        for color in colors:
            if color not in colorDict:
                if i >= len(symbols):
                    raise ValueError("Length of Contraction is too large for Einsum!")
                colorDict[color] = symbols[i]
                i += 1
    return colorDict

def get_einsum_substring(conCoreDict, openVariables, symbols = defaultSymbols):

    colorDict = get_colorDict([conCoreDict[key].colors for key in conCoreDict], symbols=symbols)
    coreOrder = list(conCoreDict.keys())
    colorOrder = list(colorDict.keys())
    leftString = ",".join([
        "".join([colorDict[color] for color in conCoreDict[key].colors])
        for key in coreOrder
    ])
    rightString = "".join([colorDict[color] for color in colorOrder if color in openVariables])

    return "->".join([leftString, rightString]), coreOrder, colorDict, colorOrder

def get_reorder_substring(oldColors, newColors, symbols = defaultSymbols):
    colorDict = {oldColors[i]: symbols[i] for i in range(len(oldColors))}
    return "->".join(["".join([colorDict[color] for color in oldColors]),
                      "".join([colorDict[color] for color in newColors])])



from tnreason import engine
import numpy as np

categoricalCoreSuffix = "_catCore"

def create_categorical_cores(categoricalsDict):
    """
    Creates a tensor network representing the constraints of
        * categoricalsDict: Dictionary of atom lists to each categorical variable
    """
    catCores = {}
    for catName in categoricalsDict.keys():
        catCores = {**catCores, **create_constraintCoresDict(categoricalsDict[catName], catName)}
    return catCores


def create_constraintCoresDict(atoms, catName):
    constraintCoresDict = {}
    for i, atomKey in enumerate(atoms):
        coreValues = np.zeros(shape=(len(atoms), 2))
        coreValues[:, 0] = np.ones(shape=(len(atoms)))
        coreValues[i, 0] = 0
        coreValues[i, 1] = 1
        constraintCoresDict[catName + "_" + atomKey + categoricalCoreSuffix] = engine.get_core()(
            coreValues, [catName, atomKey], name=catName + "_" + atomKey + categoricalCoreSuffix)
    return constraintCoresDict



import tensorflow as tf

from tnreason.engine import workload_to_numpy as cor
from tnreason.engine import subscript_creation as subc


class TensorFlowCore:
    def __init__(self, values, colors, name=None):
        self.values = tf.convert_to_tensor(values)
        self.colors = colors
        self.name = name

    def to_NumpyTensorCore(self):
        return cor.NumpyCore(self.values.numpy(), self.colors, self.name)


class TensorFlowContractor:
    def __init__(self, coreDict, openColors):
        self.tensorFlowCores = {
            key: TensorFlowCore(values=coreDict[key].values, colors=coreDict[key].colors, name=coreDict[key].name) for
            key in coreDict
        }
        self.openColors = openColors

    def einsum(self):
        substring, coreOrder, colorDict, colorOrder = subc.get_einsum_substring(self.tensorFlowCores, self.openColors)
        return TensorFlowCore(values=tf.einsum(substring,
                                               *[self.tensorFlowCores[key].values for key in coreOrder]
                                               ), colors=[color for color in colorOrder if color in self.openColors])



from tnreason import engine
from tnreason import encoding

import numpy as np


class Gibbs:
    """
    Implements Gibbs sampling with annealing support
        * networkCores: Main tensor network forming a Markov Network
        * importanceList, importanceColors: Reshaping the probability of networkCores, as in ALS
        * exponentiated: Whether raw probability (which can get negative) is exponentiated to be posified
    """

    def __init__(self, networkCores, importanceColors=[], importanceList=[({}, 1)], exponentiated=False,
                 contractionMethod=engine.defaultContractionMethod):
        self.networkCores = networkCores

        self.importanceColors = importanceColors
        self.importanceList = importanceList
        self.contractionMethod = contractionMethod

        self.exponentiated = exponentiated

    def ones_initialization(self, updateKeys, shapesDict, colorsDict=None):
        """
        Initialize all cores to be sampled with trivial cores
        """
        if colorsDict is None:
            colorsDict = {key: key for key in updateKeys}
        for updateKey in updateKeys:
            if updateKey in self.networkCores.keys():
                print("Warning: Existing core {} has been reinitialized in Gibbs!".format(updateKey))
            self.networkCores[updateKey] = encoding.create_trivial_core(updateKey, shapesDict[updateKey],
                                                                        colorsDict[updateKey])

    def annealed_sample(self, updateKeys, annealingPattern=[(10, 1)]):
        """
        Samples based on an annealing pattern.
        """
        for rep, (sweepNum, temperature) in enumerate(annealingPattern):
            sample = self.gibbs_sample(updateKeys=updateKeys, sweepNum=sweepNum, temperature=temperature)
        return sample

    def gibbs_sample(self, updateKeys, sweepNum=10, temperature=1):
        """
        Sample created to constant temperature.
        """
        positions = np.empty(shape=(sweepNum, len(updateKeys)))
        for sweep in range(sweepNum):
            for i, updateKey in enumerate(updateKeys):
                positions[sweep, i] = self.sample_core(updateKey, temperature=temperature)

        return {updateKeys[i]: positions[-1, i] for i in range(len(updateKeys))}

    def sample_core(self, updateKey, temperature):
        """
        Update a color core based on tempered marginal probability
        """
        tbUpdated = self.networkCores.pop(updateKey)
        updateColors = tbUpdated.colors
        updateShape = tbUpdated.values.shape

        updateDistribution = engine.contract({**self.networkCores, **self.importanceList[0][0],
                                              "trivialCore": encoding.create_trivial_core("trivialCore", updateShape,
                                                                                          updateColors)},
                                             openColors=updateColors,
                                             method=self.contractionMethod).multiply(self.importanceList[0][1])
        updateDistribution.reorder_colors(updateColors)
        for importanceCores, weight in self.importanceList[1:]:
            updateDistribution = updateDistribution.sum_with(
                engine.contract({**self.networkCores, **importanceCores,
                                 "trivialCore": encoding.create_trivial_core("trivialCore", updateShape,
                                                                             updateColors)},
                                openColors=updateColors,
                                method=self.contractionMethod).multiply(weight))

        flattened = updateDistribution.values.flatten()
        if self.exponentiated:  # Used to posify the unnormalized probability
            flattened = np.exp(flattened)

        if np.sum(flattened) <= 0:  # Sample from uniform distribution, when vanishing unnormalized probability
            localProb = np.ones(shape=flattened.shape) / flattened.shape[0]
            print("Vanishing prob for update of core {}!".format(updateKey))
        else:
            if temperature != 1:
                localProb = flattened ** temperature / (np.sum(flattened ** temperature))
            else:
                localProb = flattened / np.sum(flattened)

        # Draw a basis vector from the distribution
        pos = np.where(np.random.multinomial(1, localProb) == 1)[0][0]
        self.networkCores[updateKey] = encoding.create_basis_core(updateKey, updateShape, updateColors, pos)
        return pos



from tnreason import engine
from tnreason.encoding import truth_tables as ttab
from tnreason.encoding import formulas_to_cores as enform

import numpy as np

connectiveSelColorSuffix = "_actVar"
connectiveSelCoreSuffix = "_actCore"

candidatesColorSuffix = "_selVar"
candidatesCoreSuffix = "_selCore"

posPrefix = "p"


def create_architecture(neuronDict, headNeurons=[]):
    """
    Creates a tensor network of neuron cores with selection colors
        * neuronDict: Dictionary specifying to each neuronName a list of candidates (for the connective and the arguments)
        * headNeurons: List of neuronNames to be associated with hard headCores
    """
    architectureCores = {}
    for neuronName in neuronDict.keys():
        architectureCores = {**architectureCores,
                             **create_neuron(neuronName, neuronDict[neuronName][0], {
                                 posPrefix + str(i): posCandidates for i, posCandidates in
                                 enumerate(neuronDict[neuronName][1:])
                             })}
    for headNeuron in headNeurons:
        architectureCores = {**architectureCores, **enform.create_head_core(headNeuron, headType="truthEvaluation")}
    return architectureCores


def create_neuron(neuronName, connectiveList, candidatesDict={}):
    """
    Creates the cores to one neuron 
        * neuronName: String to use as prefix of the key to each core
        * connectiveList: List of connectives to be selected
        * candidatesDict: Dictionary of lists of candidates to each argument of the neuron
    """
    neuronCores = {
        neuronName + connectiveSelCoreSuffix: create_connective_selectors(neuronName, candidatesDict.keys(), connectiveList)}
    for candidateKey in candidatesDict:
        neuronCores = {**neuronCores, **create_variable_selectors(
            neuronName, candidateKey, candidatesDict[candidateKey])}
    return neuronCores


def create_variable_selectors(neuronName, candidateKey, variables):
    """
    Creates the selection cores to one argument at a neuron
    """
    cSelectorDict = {}
    for i, variableKey in enumerate(variables):
        values = np.ones(
            shape=(len(variables), 2, 2))  # control, atom (subexpression), formulatruth (headexpression)
        values[i, 0, 1] = 0
        values[i, 1, 0] = 0

        cSelectorDict[neuronName + "_" + candidateKey + "_" + variableKey + candidatesCoreSuffix] = engine.get_core()(
            values,
            [neuronName + "_" + candidateKey + candidatesColorSuffix,
             variableKey,
             candidateKey],
            neuronName + "_" + candidateKey + "_" + variableKey + candidatesCoreSuffix
        )
    return cSelectorDict


def create_connective_selectors(neuronName, candidateKeys, connectiveList):
    """
    Creates the connective selection core, using the candidateKeys as color and arity specification
    """
    if len(candidateKeys) == 1:
        values = np.empty((len(connectiveList), 2, 2))
        for i, connectiveKey in enumerate(connectiveList):
            values[i] = ttab.get_unary_tensor(connectiveKey)

    elif len(candidateKeys) == 2:
        values = np.empty((len(connectiveList), 2, 2, 2))
        for i, connectiveKey in enumerate(connectiveList):
            values[i] = ttab.get_binary_tensor(connectiveKey)
    else:
        raise ValueError(
            "Number {} of candidates wrong in Neuron {} with connectives {}!".format(len(candidateKeys), neuronName,
                                                                                     connectiveList))
    return engine.get_core()(values, [neuronName + connectiveSelColorSuffix, *candidateKeys, neuronName])


## Functions to identify solution expressions when candidates are selected
def create_solution_expression(neuronDict, selectionDict):
    """
    Replaces the candidates of neurons by solutions and returns the identified head neurons as formulas
        * neuronDict: Dictionary specifying the neurons
        * selectionDict: Dictionary selecting candidates (connective and position) to each selection variables at each neuron
    """
    fixedNeurons = fix_neurons(neuronDict, selectionDict)
    headNeurons = get_headKeys(fixedNeurons)
    if len(headNeurons) != 1:
        print("WARNING: Headneurons are {}.".format(headNeurons))
    return {headKey: replace_neuronnames(headKey, fixedNeurons) for headKey in headNeurons}


def fix_neurons(neuronDict, selectionDict):
    """
    Replaces the neurons with subexpressions refering to each other
    """
    rawFormulas = {}
    for neuronName in neuronDict:
        rawFormulas[neuronName] = [neuronDict[neuronName][0][selectionDict[neuronName + connectiveSelColorSuffix]]] + [
            neuronDict[neuronName][i][selectionDict[neuronName + "_" + posPrefix + str(i - 1) + candidatesColorSuffix]]
            for i in range(1, len(neuronDict[neuronName]))]
    return rawFormulas


def get_headKeys(fixedNeurons):
    """
    Identifies the independent neurons as heads
    """
    headKeys = set(fixedNeurons.keys())
    for formulaKey in fixedNeurons:
        for inNeuron in fixedNeurons[formulaKey][1:]:
            if inNeuron in headKeys:
                headKeys.remove(inNeuron)
    return headKeys


def replace_neuronnames(currentNeuronName, fixedNeurons):
    """
    Replaces the current neuron with the respective expression, after iterative replacement of depending fixed neurons
    """
    if currentNeuronName not in fixedNeurons:
        return currentNeuronName  ## Then an atom
    currentNeuron = fixedNeurons[currentNeuronName].copy()
    currentNeuron = [currentNeuron[0]] + [replace_neuronnames(currentNeuron[i], fixedNeurons) for i in
                                          range(1, len(currentNeuron))]
    return currentNeuron


## Auxiliary functions for knowledge identifying the atoms and the dimension of selection variables
def find_atoms(specDict):
    atoms = set()
    for neuronName in specDict.keys():
        for positionList in specDict[neuronName][1:]:
            atoms = atoms | set(positionList)
    return list(atoms)


def find_selection_dimDict(specDict):
    dimDict = {}
    for neuronName in specDict:
        dimDict.update({neuronName + connectiveSelColorSuffix: len(specDict[neuronName][0]),
                        **{neuronName + "_" + posPrefix + str(i) + candidatesColorSuffix: len(candidates)
                           for i, candidates in enumerate(specDict[neuronName][1:])}})
    return dimDict



from tentris import tentris, Hypertrie

from tnreason.engine import workload_to_numpy as cor

from tnreason.engine import subscript_creation as subc

import numpy as np


class TentrisCore:
    def __init__(self, values, colors, name=None, inType="Hypertrie"):
        if inType == "Hypertrie":
            self.values = values
        elif inType == "Numpy":
            self.values_from_numpy(array=values)
        elif inType == "rdf":
            self.values_from_rdf(path=values)

        self.colors = colors
        self.name = name

    def values_from_rdf(self, path):
        tStore = tentris.TripleStore()
        tStore.load_rdf_data(path)
        self.values = tStore.hypertrie()

    def values_from_numpy(self, array):
        self.values = Hypertrie(dtype=array.dtype, depth=len(array.shape))
        for index, coordinate in enumerate(np.nditer([array, array.shape])):
            if coordinate != 0:
                self.values[np.unravel_index(index, array.shape)] = coordinate

    def values_to_numpy(self):
        depth = self.values.depth

        ## Get size -> Better to extend to coordinate dictionaries?
        size = np.zeros(depth)
        for entry in self.values:
            for i in range(depth):
                if entry[0][i] > size[i]:
                    size[i] = entry[0][i]

        numpyValues = np.empty(shape=[int(size[i]) for i in range(depth)])
        for entry in self.values:
            numpyValues[entry[0]] = entry[1]
        return numpyValues

    def to_NumpyTensorCore(self):
        return cor.NumpyCore(self.values_to_numpy(), self.colors, self.name)


class TentrisContractor:
    def __init__(self, coreDict, openColors):
        self.tentrisCores = {
            key: TentrisCore(values=coreDict[key].values, colors=coreDict[key].colors, name=coreDict[key].name, inType="Hypertrie") for
            key in coreDict
        }
        self.openColors = openColors

    def einsum(self):
        substring, coreOrder, colorDict, colorOrder = subc.get_einsum_substring(self.tentrisCores, self.openColors)
        with tentris.einsumsum(subscript=substring, operands=[self.tentrisCores[key].values for key in coreOrder]) as e:
            resultValues = Hypertrie(dtype=e.result_dtype, depth=e.result_depth)
            e.try_extend_hypertrie(resultValues)

        return TentrisCore(values=resultValues,
                           colors=[color for color in colorOrder if color in self.openColors],
                           inType="Hypertrie")

if __name__ == "__main__":
    testCore1 = TentrisCore(np.array([[1.1, 2], [0.12, -1.1]]), colors=["a", "b"])
    testCore2 = TentrisCore(np.array([[1.1, 2], [0.12, -1.1]]), colors=["a", "c"])

    testContractor = TentrisContractor({"1": testCore1, "2":testCore2}, ["b","c"])
    print(testContractor.einsum().to_NumpyTensorCore().values)


class SliceCore:
    def __init__(self, values, colors, name=None):
        self.values = values
        self.colors = colors
        self.name = name

    def __str__(self):
        return "## Sliced Core " + str(self.name) + " ##\nValues: " + str(self.values) + "\nColors: " + str(self.colors)


class SliceContractor:
    def __init__(self, coreDict={}, openColors=[]):
        self.coreDict = coreDict
        self.openColors = openColors

    def contract(self):
        allColors = set()
        for key in self.coreDict:
            allColors = allColors | set(self.coreDict[key].colors)

        values = [(1, set(), set())]
        for key in self.coreDict:
            values = slice_contraction(values, self.coreDict[key].values)
        values = reduce_colors(values, allColors - set(self.openColors))
        return SliceCore(values, self.openColors)


def reduce_colors(values, reduceColors):
    reducedValues = []
    for val, pos, neg in values:
        for reduceColor in reduceColors:
            if reduceColor not in pos and reduceColor not in neg:
                val = 2 * val  # Color dimension always 2
            elif reduceColor in pos:
                pos.remove(reduceColor)
            elif reduceColor in neg:
                neg.remove(reduceColor)
            else:
                print(reduceColor, pos, neg, reduceColor in pos, reduceColor in neg)
        reducedValues.append((val, pos, neg))
    return reducedValues


def slice_contraction(values1, values2):
    slices = []
    for slice1 in values1:
        for slice2 in values2:
            combined = combine_slices(slice1, slice2)
            if combined[0] != 0:
                slices.append(combined)
    return slices


def combine_slices(slice1, slice2):
    val1, pos1, neg1 = slice1
    val2, pos2, neg2 = slice2

    pos = pos1 | pos2
    neg = neg1 | neg2
    if len(pos & neg) > 0:  ## Check whether the slice zero
        return (0, [], [])
    else:
        return (val1 * val2, pos, neg)


if __name__ == "__main__":
    values1 = [
        (1.1, {"a", "b"}, {"c"}),
        (0.9, set(), {"d"})
    ]
    core1 = SliceCore(values1, ["a", "b", "c"])

    values2 = [
        (1.1, {"b"}, {"a", "c"}),
        (2, {"a"}, set())
    ]
    core2 = SliceCore(values2, ["a", "b", "c"])

    contracted = SliceContractor(coreDict={
        "c1": core1,
        "c2": core2
    }, openColors=["a", "b"]).contract()

    print(contracted)

    # print([e for e in slice_contraction(values1, values2)])



from tnreason import engine
from tnreason import encoding

import numpy as np


class ALS:
    """
    Implements the alternating least squares
        * networkCores: Main tensor network to be optimized
        * importanceList: List of tuples containing
            - tensor network specifying a loss by contraction
            - weight specifying the importance in the loss
        * importanceColors: Specifying the shared colors of networkCores and importanceList networks
        * targetCores: Specifying the fitting target after contraction
        * trivialKeys: Specifying cores of singe coordinates, which contribute only factors
    """

    def __init__(self, networkCores, importanceColors=[], importanceList=[({}, 1)],
                 contractionMethod=engine.defaultContractionMethod, targetCores={}):
        self.networkCores = networkCores

        self.importanceColors = importanceColors
        self.importanceList = importanceList
        self.contractionMethod = contractionMethod

        self.targetCores = targetCores

        self.trivialKeys = []  # Keys with single position, trivial in the sense that they will not be updated

    def random_initialize(self, updateKeys, shapesDict={}, colorsDict={}):
        for updateKey in updateKeys:
            if updateKey in self.networkCores:
                upShape = self.networkCores[updateKey].values.shape
                upColors = self.networkCores[updateKey].colors
                self.networkCores.pop(updateKey)
            else:
                upShape = shapesDict[updateKey]
                upColors = colorsDict[updateKey]
            if np.prod(upShape) > 1:
                self.networkCores[updateKey] = encoding.create_random_core(updateKey, upShape, upColors,
                                                                           randomEngine="NumpyUniform")
            else:
                self.trivialKeys.append(updateKey)
                self.networkCores[updateKey] = encoding.create_trivial_core(updateKey, upShape, upColors)

    def alternating_optimization(self, updateKeys, sweepNum=10, computeResiduum=False):
        updateKeys = [key for key in updateKeys if key not in self.trivialKeys]
        if computeResiduum:
            residua = np.empty((sweepNum, len(updateKeys)))
        for sweep in range(sweepNum):
            for i, updateKey in enumerate(updateKeys):
                self.optimize_core(updateKey)
                if computeResiduum:
                    residua[sweep, i] = self.compute_residuum()
        if computeResiduum:
            return residua

    def get_color_argmax(self, updateKeys):
        # ! Only working for vectors #
        return {self.networkCores[key].colors[0]: np.argmax(np.abs(self.networkCores[key].values)) for key in
                updateKeys}

    def optimize_core(self, updateKey):
        ## Trivialize the core to be updated (serving as a placeholder)
        tbUpdated = self.networkCores.pop(updateKey)
        self.networkCores[updateKey] = encoding.create_trivial_core(updateKey, tbUpdated.values.shape, tbUpdated.colors)

        ## Compute flattened operator and target
        updateColors = tbUpdated.colors
        conOperator = self.compute_conOperator(updateColors, importanceCores=self.importanceList[0][0],
                                               weight=self.importanceList[0][1])
        conTarget = self.compute_conTarget(updateColors, importanceCores=self.importanceList[0][0],
                                           weight=self.importanceList[0][1])
        for importanceCores, weight in self.importanceList[1:]:
            conOperator = conOperator.sum_with(
                self.compute_conOperator(updateColors, importanceCores, weight))
            conTarget = conTarget.sum_with(self.compute_conTarget(updateColors, importanceCores, weight))

        resultDim = int(np.prod(conTarget.values.shape))
        conOperator.reorder_colors(conTarget.colors + [color + "_out" for color in conTarget.colors])
        flattenedOperator = conOperator.values.reshape(resultDim, resultDim)
        flattenedTarget = conTarget.values.flatten()

        ## Update the core by solution of least squares problem
        solution, res, rank, s = np.linalg.lstsq(flattenedOperator, flattenedTarget, rcond=None)
        self.networkCores[updateKey] = engine.get_core()(solution.reshape(tbUpdated.values.shape), updateColors,
                                                         updateKey)

    def compute_conOperator(self, updateColors, importanceCores={}, weight=1):
        return engine.contract(method=self.contractionMethod,
                               coreDict={
                                   **importanceCores,
                                   **self.networkCores,
                                   **copy_cores(self.networkCores, "_out", self.importanceColors)
                               }, openColors=updateColors + [updateColor + "_out" for updateColor in
                                                             updateColors]).multiply(weight)

    def compute_conTarget(self, updateColors, importanceCores={}, weight=1):
        return engine.contract(method=self.contractionMethod,
                               coreDict={
                                   **importanceCores,
                                   **self.networkCores,
                                   **self.targetCores,
                               }, openColors=updateColors).multiply(weight)

    def compute_residuum(self):
        prediction = engine.contract(method=self.contractionMethod,
                                     coreDict=self.networkCores,
                                     openColors=self.importanceColors)
        target = engine.contract(method=self.contractionMethod,
                                 coreDict=self.targetCores,
                                 openColors=self.importanceColors)
        prediction.reorder_colors(target.colors)
        ## Not using the weightings by the importanceList!
        return np.linalg.norm(prediction.values - target.values)


def copy_cores(coreDict, suffix, exceptionColors):
    returnDict = {}
    for key in coreDict:
        core = coreDict[key].clone()
        newColors = core.colors
        for i, color in enumerate(newColors):
            if color not in exceptionColors:
                newColors[i] = color + suffix
        core.colors = newColors
        returnDict[key + suffix] = core
    return returnDict



import numpy as np

from tnreason.engine import subscript_creation as subc

class NumpyCore:
    def __init__(self, values, colors, name=None):
        if len(colors) != len(values.shape):
            raise ValueError("Number of Colors does not match the Value Shape in Core {}!".format(name))
        if len(colors) != len(set(colors)):
            raise ValueError("There are duplicate colors in the colors {} of Core {}!".format(colors, name))

        self.values = values
        self.colors = colors
        self.name = name

    def get_values_as_array(self):
        return self.values

    def clone(self):
        return NumpyCore(self.values.copy(), self.colors.copy(), self.name)  # ! Shallow Copies?

    ## For Sampling
    def normalize(self):
        return NumpyCore(1 / np.sum(self.values) * self.values, self.colors, self.name)

    ## For ALS: Reorder Colors and summation
    def reorder_colors(self, newColors):
        self.values = np.einsum(subc.get_reorder_substring(self.colors, newColors), self.values)
        self.colors = newColors

    def sum_with(self, sumCore):
        if set(self.colors) != set(sumCore.colors):
            print(self.colors, sumCore.colors)
            raise ValueError("Colors of summands {} and {} do not match!".format(self.name, sumCore.name))
        else:
            self.reorder_colors(sumCore.colors)
            return NumpyCore(self.values + sumCore.values, self.colors, self.name)

    def multiply(self, weight):
        return NumpyCore(weight * self.values, self.colors, self.name)

    def get_maximal_index(self):
        return np.unravel_index(np.argmax(self.values.flatten()), self.values.shape)


class NumpyEinsumContractor:
    def __init__(self, coreDict={}, openColors=[]):
        self.coreDict = {key: coreDict[key].clone() for key in coreDict}
        self.openColors = openColors

    def contract(self):
        substring, coreOrder, colorDict, colorOrder = subc.get_einsum_substring(self.coreDict, self.openColors)
        return NumpyCore(
            np.einsum(substring, *[self.coreDict[key].values for key in coreOrder]),
            [color for color in colorOrder if color in self.openColors])


from tnreason import engine
from tnreason import encoding

import numpy as np

momentCoreSuffix = encoding.headCoreSuffix
targetCoreSuffix = "_targetCore"


class MomentMatcher:
    """
    Fits alternatingly local cores to reproduce local expected Statistics.
    Equals to coordinate descent with optimal steplength of the likelihood.
        * targetCores: Vector Cores storing the local expected statistics to be matched
        * networkCores: Static Cores shaping the basis
    Generalizes the weight estimation (which is the special case of leg dimension 2 and fitting of exponentiated first coordinate.
    """

    def __init__(self, networkCores, targetCores):
        self.networkCores = networkCores

        self.targetCores = {targetCores[key].colors[0] + targetCoreSuffix: targetCores[key] for key in targetCores}

        self.updateDimDict = {self.targetCores[key].colors[0]: self.targetCores[key].values.shape[0] for key in
                              self.targetCores}

    def ones_initialization(self):
        """
        varDimDict: Dictionary with keys the colors of the moments and the shape the dimension of the axis
        """
        self.networkCores.update(
            encoding.create_trivial_cores(list(self.updateDimDict.keys()), shapeDict=self.updateDimDict,
                                          suffix=momentCoreSuffix)
        )

    def matching_step(self, updateColor):
        self.networkCores.pop(updateColor + momentCoreSuffix)

        self.networkCores[updateColor + momentCoreSuffix] = engine.get_core()(
            values=solve_moment_equation(
                satVect=engine.contract(coreDict=self.networkCores, openColors=[updateColor]).values,
                empVect=self.targetCores[updateColor + targetCoreSuffix].values
            ), colors=[updateColor], name=updateColor + momentCoreSuffix)

        print(self.networkCores[updateColor + momentCoreSuffix].values)

    def alternating_matching(self, sweepNum=10, updateColors=None):
        if updateColors is None:
            updateColors = list(self.updateDimDict.keys())
        for sweep in range(sweepNum):
            for updateColor in updateColors:
                self.matching_step(updateColor)


def find_common_nonzero(vect1, vec2):
    """
    Searching for a reference position for quotients in moment match.
    """
    for i in range(vect1.shape[0]):
        if vect1[i] > 0 and vec2[i] > 0:
            return i
    else:
        return -1


def solve_moment_equation(satVect, empVect):
    """
    Solves the local expected statistics matching of
        * satVect: Marginal probability of color wrt alien cores
        * empVect: Desired marginal probability (expected statistics)
    """
    solVect = np.ones(shape=satVect.shape)

    refPos = find_common_nonzero(satVect, empVect)
    if refPos == -1:
        print("Warning: Moments cannot be matched!")
        return solVect

    for i in range(satVect.shape[0]):
        solVect[i] = (satVect[refPos] / empVect[refPos]) * (empVect[i] / satVect[i])

    return solVect



from tnreason import engine

import numpy as np

from tnreason.encoding import truth_tables as ttab

connectiveFixCoreSuffix = "_conCore"
headCoreSuffix = "_headCore"


def create_formulas_cores(expressionsDict, alreadyCreated=[]):
    """
    Creates a tensor network of connective and head cores
        * expressionsDict: Dictionary of nested listed representing expressions
        * alreadyCreated: List of keys to connective cores to be omitted
    """
    knowledgeCores = {}
    for formulaName in expressionsDict.keys():
        if isinstance(expressionsDict[formulaName][-1], float) or isinstance(expressionsDict[formulaName][-1], int):
            knowledgeCores = {**knowledgeCores,
                              **create_head_core(get_formula_color(expressionsDict[formulaName][:-1]), "expFactor", weight=
                              expressionsDict[formulaName][-1]),
                              **create_raw_formula_cores(expressionsDict[formulaName][:-1],
                                                         alreadyCreated=
                                                         list(knowledgeCores.keys()) + alreadyCreated)}
        else:
            knowledgeCores = {**knowledgeCores,
                              **create_head_core(get_formula_color(expressionsDict[formulaName]), "truthEvaluation"),
                              **create_raw_formula_cores(expressionsDict[formulaName],
                                                         alreadyCreated=list(knowledgeCores.keys()) + alreadyCreated)}
    return knowledgeCores


def create_raw_formula_cores(expression, alreadyCreated=[]):
    """
    Creates the connective cores to an expression, omitting the elsewhere created cores
        * expression: Nested list specifying a formula
        * alreadyCreated: List of keys to connective cores to be omitted
    """
    if get_formula_color(expression) + connectiveFixCoreSuffix in alreadyCreated:
        return {}
    if isinstance(expression, str):
        return {}
    elif len(expression) == 1:
        assert isinstance(expression[0], str)
        return {}

    elif len(expression) == 2:
        return {**create_connective_core(expression),
                **create_raw_formula_cores(expression[1], alreadyCreated=alreadyCreated)}
    elif len(expression) == 3:
        return {**create_connective_core(expression),
                **create_raw_formula_cores(expression[1], alreadyCreated=alreadyCreated),
                **create_raw_formula_cores(expression[2], alreadyCreated=alreadyCreated)
                }
    else:
        raise ValueError("Expression {} not understood!".format(expression))


def create_connective_core(expression):
    """
    Creates the connective core at the head of the expression by loading the truth table
    """
    expressionString = get_formula_color(expression)
    if isinstance(expression, str):
        return {}

    elif len(expression) == 2:
        preExpressionString = get_formula_color(expression[1])
        return {expressionString + connectiveFixCoreSuffix: engine.get_core()(
            ttab.get_unary_tensor(expression[0]),
            [preExpressionString, expressionString],
            expressionString + connectiveFixCoreSuffix)
        }

    elif len(expression) == 3:
        leftExpressionString = get_formula_color(expression[1])
        rightExpressionString = get_formula_color(expression[2])
        return {
            expressionString + connectiveFixCoreSuffix: engine.get_core()(
                ttab.get_binary_tensor(expression[0]),
                [leftExpressionString,
                 rightExpressionString,
                 expressionString],
                expressionString + connectiveFixCoreSuffix)}
    else:
        raise ValueError("Expression {} not understood!".format(expression))


def create_head_core(expression, headType, weight=None, name=None):
    """
    Created the head core to an expression activating it
    """
    if headType == "truthEvaluation":
        headValues = np.zeros(shape=(2))
        headValues[1] = 1
    elif headType == "falseEvaluation":
        headValues = np.zeros(shape=(2))
        headValues[0] = 1
    elif headType == "weightedTruthEvaluation":
        headValues = np.zeros(shape=(2))
        headValues[1] = weight
    elif headType == "expFactor":
        headValues = np.zeros(shape=(2))
        headValues[0] = 1
        headValues[1] = np.exp(weight)
    elif headType == "diffExpFactor":
        headValues = np.zeros(shape=(2))
        headValues[1] = np.exp(weight)
    else:
        raise ValueError("Headtype {} not understood!".format(headType))

    color = get_formula_color(expression)

    if name is None:
        name = color + headCoreSuffix

    return {name: engine.get_core()(headValues, [color], name)}

def create_evidence_cores(evidenceDict):
    """
    Turns positive and negative evidence into literal formulas and encodes them
    """
    return create_formulas_cores({**{key: [key] for key in evidenceDict if evidenceDict[key]},
                                  **{key: ["not", key] for key in evidenceDict if not evidenceDict[key]}
                                  })

def get_formula_color(expression):
    """
    Identifies a color with an expression
    """
    if isinstance(expression, str):  ## Expression is atomic
        return expression
    elif len(expression) == 1:  ## Expression is atomic, but provided in nested form
        assert isinstance(expression[0], str)
        return expression[0]
    else:
        if not isinstance(expression[0], str):
            raise ValueError("Connective {} has wrong type!".format(expression[0]))
        return "(" + expression[0] + "_" + "_".join(
            [get_formula_color(entry) for entry in expression[1:]]) + ")"

def get_all_atoms(expressionsDict):
    """
    Identifies the leafs of the expressions in the expressionsDict as atoms
    """
    atoms = set()
    for key in expressionsDict:
        atoms = atoms | get_atoms(expressionsDict[key])
    return list(atoms)


def get_atoms(expression):
    if isinstance(expression, str):  ## Then an atom
        return {expression}
    elif len(expression) == 1:  ## Then an atomic formula
        return {expression[0]}
    else:  ## Then a formula with connective in first position
        atoms = set()
        for subExpression in expression[1:]:
            atoms = atoms | get_atoms(subExpression)
        return atoms



# knowledge/__init__.py

from tnreason.knowledge.inductive import HybridLearner
from tnreason.knowledge.deductive import InferenceProvider

from tnreason.knowledge.weight_estimation import EntropyMaximizer
from tnreason.knowledge.distributions import HybridKnowledgeBase, EmpiricalDistribution
from tnreason.knowledge.formula_boosting import FormulaBooster
from tnreason.knowledge.batch_evaluation import KnowledgePropagator

from tnreason.knowledge.knowledge_visualization import visualize

def load_kb_from_yaml(loadPath):
    kb = HybridKnowledgeBase()
    kb.from_yaml(loadPath)
    return kb





import networkx as nx
from matplotlib import pyplot as plt

from tnreason import encoding


def visualize(hybridKB, title="Visualization of the Knowledge Base",
              savePath=None):
    expressionsDict = hybridKB.weightedFormulas
    factsDict = hybridKB.facts
    categoricalConstraints = hybridKB.categoricalConstraints
    evidenceDict = hybridKB.evidence

    graph, pos, atoms = visualize_subexpressions(expressionsDict=expressionsDict, factsDict=factsDict,
                                                 categoricalConstraints=categoricalConstraints,
                                                 evidenceDict=evidenceDict)
    draw_with_evidence(graph, pos, evidenceDict, atoms, title=title, savePath=savePath)


def visualize_with_differing_evidence(expressionsDict={},
                                      factsDict={},
                                      categoricalConstraints={},
                                      evidenceList=[],
                                      title="Visualization of the Knowledge Base",
                                      saveBasePath="../../tests/visualization"):
    graph, pos, atoms = visualize_subexpressions(expressionsDict=expressionsDict, factsDict=factsDict,
                                                 categoricalConstraints=categoricalConstraints,
                                                 evidenceDict={})
    for i, evidenceDict in enumerate(evidenceList):
        draw_with_evidence(graph, pos, evidenceDict, atoms, title=title, savePath=saveBasePath + str(i) + ".png")


def get_edges_and_positions(expression):
    expressionString = encoding.get_formula_color(expression)

    if isinstance(expression, str):
        return [], {expressionString: 0}, [expressionString]
    else:
        edges, positions, subExpressions = [], {}, [expressionString]
        for subExpression in expression:
            if is_subexpression(subExpression):
                subEdges, subPositions, subSubExpressions = get_edges_and_positions(subExpression)
                edges = edges + subEdges
                positions = {**positions, **subPositions}
                subExpressions = subExpressions + subSubExpressions

                edges.append([encoding.get_formula_color(expression), encoding.get_formula_color(subExpression)])

        positions[encoding.get_formula_color(expression)] = 1 + max(
            [positions[encoding.get_formula_color(subExpression)] for subExpression in expression if
             is_subexpression(subExpression)])

        return edges, positions, subExpressions


def visualize_subexpressions(expressionsDict={},
                             factsDict={},
                             categoricalConstraints={},
                             evidenceDict={}):
    edges = []
    horPositions = {}
    nodes = []

    allExpressions = list(expressionsDict.values()) + list(factsDict.values()) + [[key] for key in evidenceDict]
    for key in categoricalConstraints:
        allExpressions = allExpressions + categoricalConstraints[key]

    for expression in allExpressions:
        expressionEdges, expressionPositions, subexpressions = get_edges_and_positions(expression)
        edges.extend(expressionEdges)
        horPositions = {
            **horPositions,
            **expressionPositions}
        nodes.extend(subexpressions)

    graph = nx.Graph()
    graph.add_edges_from(edges)
    graph.add_nodes_from(nodes)

    pos = nx.spring_layout(graph, k=0.8)
    for nodeKey in pos:
        pos[nodeKey][0] = horPositions[nodeKey]

    atoms = encoding.get_all_atoms({**{key: expressionsDict[key][:-1] for key in expressionsDict}, **factsDict})
    return graph, pos, atoms


def draw_with_evidence(graph, pos, evidenceDict, atoms, title, savePath=None):
    trueColor = "blue"
    falseColor = "red"
    neutralColor = "gray"

    hidden_node_alpha = 0.3
    visible_node_alpha = 0.8

    nx.draw_networkx_nodes(graph, pos,
                           nodelist=[atomKey for atomKey in evidenceDict if
                                     evidenceDict[atomKey] and atomKey not in atoms],
                           node_size=1000,
                           node_color=trueColor,
                           alpha=hidden_node_alpha)
    nx.draw_networkx_nodes(graph, pos,
                           nodelist=[atomKey for atomKey in evidenceDict if
                                     not evidenceDict[atomKey] and atomKey not in atoms],
                           node_size=1000,
                           node_color=falseColor,
                           alpha=hidden_node_alpha)
    nx.draw_networkx_nodes(graph, pos,
                           nodelist=[atomKey for atomKey in graph.nodes if
                                     atomKey not in evidenceDict and atomKey not in atoms],
                           node_size=1000,
                           node_color=neutralColor,
                           alpha=hidden_node_alpha)

    ## Visible
    nx.draw_networkx_nodes(graph, pos,
                           nodelist=[atomKey for atomKey in evidenceDict if evidenceDict[atomKey] and atomKey in atoms],
                           node_size=1000,
                           node_color=trueColor,
                           alpha=visible_node_alpha)
    nx.draw_networkx_nodes(graph, pos,
                           nodelist=[atomKey for atomKey in evidenceDict if
                                     not evidenceDict[atomKey] and atomKey in atoms],
                           node_size=1000,
                           node_color=falseColor,
                           alpha=visible_node_alpha)
    nx.draw_networkx_nodes(graph, pos,
                           nodelist=[atomKey for atomKey in atoms if atomKey not in evidenceDict and atomKey in atoms],
                           node_size=1000,
                           node_color=neutralColor,
                           alpha=visible_node_alpha)

    nx.draw_networkx_labels(graph, pos, {nodeKey: get_symbol(nodeKey) for nodeKey in graph.nodes})
    edges = [edge for edge in graph.edges if edge[0] != edge[1]]
    nx.draw_networkx_edges(graph, pos, edges)

    plt.title(title, fontsize=15)
    if savePath is not None:
        plt.savefig(savePath)
    plt.show()

def is_subexpression(expression):
    if type(expression) == list:
        return True
    elif type(expression) == str:
        return not expression in ["id", "not", "and", "or", "xor", "imp", "eq"]
    else:
        return False


def get_symbol(expressionString):
    connectiveDict = {
        "and": "and",
        "not": "not"
    }
    if not "_" in expressionString:
        return expressionString
    else:
        connective = expressionString.split("_")[0][1:]
        if connective in connectiveDict:
            return connectiveDict[connective]
        return connective


if __name__ == "__main__":
    visualize_with_differing_evidence(
        expressionsDict={"e1": ["not", ["and", "b", "c"], 2]},
        factsDict={"f1": ["c"]},
        categoricalConstraints={"c1": ["e", "c"]},
        evidenceList=[{"c": 0}, {"c": 1, "e": 0}])



import yaml

def save_as_yaml(modelSpec, savePath):
    with open(savePath, "w") as file:
        yaml.dump(modelSpec, file)

def load_from_yaml(loadPath):
    with open(loadPath) as file:
        data = yaml.safe_load(file)
    return data


import networkx as nx
import matplotlib.pyplot as plt

def draw_contractionDiagram(coreDict, fontsize=10, title="Contraction Diagram", pos=None):
    ##

    coreNodes = list(coreDict.keys())

    colorNodes = []
    ccEdges = []
    for coreKey in coreDict:
        for color in coreDict[coreKey].colors:
            if color not in colorNodes:
                colorNodes.append(color)
            ccEdges.append([color, coreKey])

    graph = nx.DiGraph()
    graph.add_nodes_from(coreNodes)
    graph.add_nodes_from(colorNodes)
    graph.add_edges_from(ccEdges)

    if pos is None:
        pos = nx.spring_layout(graph)

    labels = {atom: atom for atom in coreNodes + colorNodes}
    nx.draw_networkx_labels(graph, pos, labels, font_size=fontsize)

    nx.draw_networkx_nodes(graph, pos,
                           nodelist=colorNodes,
                           node_color='r',
                           node_size=100,
                           alpha=0.2)

    nx.draw_networkx_nodes(graph, pos,
                           nodelist=coreNodes,
                           node_color='b',
                           node_size=300,
                           alpha=0.2)

    nx.draw_networkx_edges(graph, pos,
                           edgelist=ccEdges,
                           arrowstyle="<-",
                           width=2, alpha=1, edge_color='tab:grey')
    plt.title(title)
    plt.show()



from pgmpy.models import MarkovNetwork
from pgmpy.factors.discrete import DiscreteFactor
from pgmpy.inference import VariableElimination

from tnreason.engine import workload_to_numpy as cor

class PgmpyVariableEliminator:
    """
    Executed Contractions using the Variable Elimination Algorithm in Pgmpy.
    Outputs by default a Numpy Core
    """
    def __init__(self, coreDict={}, openColors=[]):
        self.model = MarkovNetwork()
        self.add_factors_from_coresDict(coreDict)

        self.openColors = openColors

    def add_factors_from_coresDict(self, coresDict):
        for coreKey in coresDict:
            for col1 in coresDict[coreKey].colors:
                self.model.add_node(col1)
                for col2 in coresDict[coreKey].colors:
                    if col2 != col1:
                        self.model.add_edge(col1, col2)
            self.model.add_factors(
                DiscreteFactor(coresDict[coreKey].colors, coresDict[coreKey].values.shape, coresDict[coreKey].values)
            )

    def contract(self):
        result = VariableElimination(self.model).query(evidence={}, variables=self.openColors)
        return cor.NumpyCore(result.values, result.variables)


from tnreason import engine
from tnreason import encoding

import numpy as np

answerCoreSuffix = "_answerCore"


class BinaryContractionPropagator:
    """
    Generalizing the Constraint Propagation towards propagated generic tensor cores (whereas ConstraintPropagator propagates vector cores)
        Avoid multiple elements in queue -> Usage of sets
    """

    def __init__(self, subNetworks, queryVariables, affectionDict=None):
        self.subNetworks = subNetworks

        self.queryVariables = queryVariables

        self.initialize_answerCores(queryVariables, find_all_color_shapes(subNetworks))

        if affectionDict is None:
            affectionDict = find_affection_by_colors(subNetworks, queryVariables)
        self.affectionDict = affectionDict

    def initialize_answerCores(self, queryVariables, colorShapes):
        if not all([all([color in colorShapes for color in queryVariables[key]]) for key in queryVariables]):
            raise ValueError("A query color is not appearing in the subnetworks!")

        self.answerCores = {key + answerCoreSuffix: engine.get_core()(
            values=np.ones(shape=[colorShapes[color] for color in queryVariables[key]]),
            colors=queryVariables[key]) for key in queryVariables}

    def initialize_queue(self):
        self.networkQueue = set()
        for queryKey in self.affectionDict:
            for netKey in self.affectionDict[queryKey]:
                self.networkQueue.add(netKey)

    def propagate(self):
        while not len(self.networkQueue) == 0:
            networkKey = self.networkQueue.pop()
            for queryKey in [key for key in self.queryVariables if networkKey in self.affectionDict[key]]:
                changed = self.update_answerCore(queryKey, networkKey)
                if changed:
                    for netKey in self.affectionDict[queryKey]:
                        self.networkQueue.add(netKey)

    def update_answerCore(self, queryKey, networkKey):
        contracted = engine.contract(coreDict={**self.subNetworks[networkKey],
                                               **{key + answerCoreSuffix: self.answerCores[key + answerCoreSuffix] for
                                                  key in self.affectionDict if
                                                  networkKey in self.affectionDict[key]}
                                               },
                                     openColors=self.queryVariables[queryKey])
        contractedShape = contracted.values.shape
        changed = False
        for i, coordinate in enumerate(contracted.values.flatten()):
            if self.answerCores[queryKey + answerCoreSuffix].values[np.unravel_index(i, contractedShape)] == 1 and \
                    contracted.values[np.unravel_index(i, contractedShape)] == 0:
                self.answerCores[queryKey + answerCoreSuffix].values[np.unravel_index(i, contractedShape)] = 0
                changed = True

        return changed


def find_affection_by_colors(subNetworks, queryVariables):
    subColors = {key: find_color_shapes(subNetworks[key]) for key in subNetworks}
    return {queryKey: [key for key in subNetworks if
                       any([color in subColors[key] for color in queryVariables[queryKey]])] for queryKey in
            queryVariables}


def find_color_shapes(network):
    colorShapes = {}
    for coreKey in network:
        for i, color in enumerate(network[coreKey].colors):
            colorShapes.update({color: network[coreKey].values.shape[i]})
    return colorShapes


def find_all_color_shapes(subNetworks):
    colorShapes = {}
    for netKey in subNetworks:
        colorShapes.update(find_color_shapes(subNetworks[netKey]))
    return colorShapes


if __name__ == "__main__":
    subNetworks = {"net1": encoding.create_formulas_cores({"f1": ["imp", "a", "b"]}),
                   "net2": encoding.create_formulas_cores({"f2": ["a"],
                                                           "f3": ["imp", "a", "c"]})}

    lCon = BinaryContractionPropagator(subNetworks, queryVariables={"q1": ["a"],
                                                                    "q3": ["c"],
                                                                    "q2": ["b", "c"]})

    lCon.initialize_queue()

    lCon.propagate()
    print(lCon.answerCores["q2" + answerCoreSuffix].values)



"""
TNReason: A package for reasoning with Tensor Networks with the subpackages
* engine: Provides the bindings to tensor network contractors
* encoding: Represents maps by tensors
* algorithms: Implements general purpose optimization algorithms on tensor networks
* knowledge: Executes deductive and inductive reasoning tasks based on markov logic networks
"""

__version__ = "1.0.0"



from tnreason import engine
from tnreason import encoding

import numpy as np


class EntropyMaximizer:
    """
    Optimizes the weights of weighted formulas based on an entropy maximization principle.
    Independent implementation of the special case of two-dimensional exponentiated weights of algorithms.MomentMatcher
    """

    def __init__(self, expressionsDict, satisfactionDict, backCores={},
                 contractionMethod=engine.defaultContractionMethod):
        self.backCores = backCores
        self.expressionsDict = expressionsDict
        self.satisfactionDict = satisfactionDict
        for key in self.satisfactionDict:
            assert self.satisfactionDict[key] <= 1 and self.satisfactionDict[
                key] >= 0, "Empirical satisfaction rate {} of key {} is wrong.".format(self.satisfactionDict[key], key)
        self.formulaCores = encoding.create_formulas_cores({
            key: expressionsDict[key] + [0] for key in expressionsDict
        })

        self.contractionMethod = contractionMethod

    def alternating_optimization(self, sweepNum=10, updateKeys=None):
        factsDict = {}
        if updateKeys is None:
            updateKeys = list(self.satisfactionDict.keys())
        for optKey in updateKeys:
            if self.satisfactionDict[optKey] == 0:
                updateKeys.remove(optKey)
                self.formulaCores.update(
                    encoding.create_head_core(self.expressionsDict[optKey], headType="falseEvaluation"))
                print("Formula {} is never satisfied.".format(self.expressionsDict[optKey]))
                factsDict[optKey] = 0
            elif self.satisfactionDict[optKey] == 1:
                updateKeys.remove(optKey)
                self.formulaCores.update(
                    encoding.create_head_core(self.expressionsDict[optKey], headType="truthEvaluation"))
                print("Formula {} is always satisfied.".format(self.expressionsDict[optKey]))
                factsDict[optKey] = 1
        weightDict = {key: [] for key in updateKeys}
        for sweep in range(sweepNum):
            for optKey in updateKeys:
                local_weight = self.local_condition_satisfier(optKey, self.satisfactionDict[optKey])
                weightDict[optKey].append(local_weight)
                self.formulaCores.update(encoding.create_head_core(self.expressionsDict[optKey], headType="expFactor",
                                                                   weight=local_weight))
        return weightDict, factsDict

    def local_condition_satisfier(self, optKey, empRate):
        optColor = encoding.get_formula_color(self.expressionsDict[optKey])
        negValue, posValue = engine.contract(method=self.contractionMethod,
                                             coreDict={**self.backCores,
                                                       **{key: self.formulaCores[key] for key in self.formulaCores if
                                                          key != optColor + encoding.headCoreSuffix}},
                                             openColors=[optColor]).values
        if negValue != 0 and posValue != 0:
            return np.log((negValue / posValue) * (empRate / (1 - empRate)))
        elif negValue == 0 or posValue == 0:
            return 0  ## In this case the formula is redundant



from tnreason import engine
import numpy as np

dataCoreSuffix = "_dataCore"
defaultDataColor = "j"


def create_data_cores(sampleDf, atomKeys=None, dataColor=defaultDataColor):
    """
    Creates a tensor network of data cores, each storing the by atomKey selected column of sampleDf as a core of
    the CP Decomposition of the one-hot encoding (empirical distribution)
    """
    if atomKeys is None:
        atomKeys = list(sampleDf.columns[1:])
    return {
        atomKey + dataCoreSuffix: atomValues_from_sampleDf(sampleDf, atomKey, dataColor)
        for atomKey in atomKeys if atomKey in list(sampleDf.columns)}


def atomValues_from_sampleDf(sampleDf, atomKey, dataColor):
    dataNum = sampleDf.values.shape[0]
    dfEntries = sampleDf[atomKey].values
    values = np.zeros(shape=(dataNum, 2))
    for i in range(dataNum):
        values[i, 1] = dfEntries[i]
        values[i, 0] = 1 - dfEntries[i]
    return engine.get_core()(values, [dataColor, atomKey], name=atomKey + dataCoreSuffix)



from tnreason import engine

import numpy as np

from queue import Queue

domainCoreSuffix = "_domainCore"


class ConstraintPropagator:
    """
    Updates binary domain cores based on local contractions.
    """
    def __init__(self, binaryCoresDict, domainCoresDict=None, verbose=False):
        self.verbose = verbose
        self.binaryCoresDict = binaryCoresDict

        if domainCoresDict is None:
            self.initialize_domainCoresDict()
        else:
            self.domainCoresDict = domainCoresDict

        self.coreQueue = Queue()
        self.create_affectionDict()
        self.assignments = None

    def initialize_domainCoresDict(self):
        self.domainCoresDict = {}
        for coreKey in self.binaryCoresDict:
            for i, color in enumerate(self.binaryCoresDict[coreKey].colors):
                if color + domainCoreSuffix not in self.domainCoresDict:
                    self.domainCoresDict[color + domainCoreSuffix] = engine.get_core()(
                        np.ones(self.binaryCoresDict[coreKey].values.shape[i]),
                        [color],
                        color + domainCoreSuffix)

    def create_affectionDict(self):
        self.colorAffectionDict = {}
        for coreKey in self.binaryCoresDict:
            for color in self.binaryCoresDict[coreKey].colors:
                if color in self.colorAffectionDict:
                    self.colorAffectionDict[color].append(coreKey)
                else:
                    self.colorAffectionDict[color] = [coreKey]

    def propagate_cores(self, coreOrder=None):
        if coreOrder is None:
            coreOrder = list(self.binaryCoresDict.keys())
        for coreKey in coreOrder:
            self.coreQueue.put(coreKey)
        while not self.coreQueue.empty():
            self.propagation_step(self.coreQueue.get())

    def propagation_step(self, coreKey):
        if self.verbose:
            print("Propagating core {}.".format(coreKey))
        changedColors = []
        for color in self.binaryCoresDict[coreKey].colors:
            contracted = engine.contract(coreDict=
                                         {coreKey: self.binaryCoresDict[coreKey],
                                          **{otherColor + domainCoreSuffix: self.domainCoresDict[
                                              otherColor + domainCoreSuffix] for otherColor in
                                             self.binaryCoresDict[coreKey].colors if otherColor != color}},
                                         openColors=[color]
                                         ).values
            for i in range(len(contracted)):
                if contracted[i] == 0 and self.domainCoresDict[color + domainCoreSuffix].values[i] == 1:
                    self.domainCoresDict[color + domainCoreSuffix].values[i] = 0
                    if color not in changedColors:
                        changedColors.append(color)
                    if np.sum(self.domainCoresDict[color + domainCoreSuffix].values) == 0:
                        raise ValueError(
                            "Inconsistent Knowledge Base: Color {} has no possible assignments!".format(color))

        for changedColor in changedColors:
            for key in self.colorAffectionDict[changedColor]:
                if key not in list(self.coreQueue.queue) and key != coreKey:
                    self.coreQueue.put(key)

    def find_assignments(self):
        self.assignments = {
            self.domainCoresDict[key].colors[0]: np.where(self.domainCoresDict[key].values == 1)[0][0]
            for key in self.domainCoresDict if np.sum(self.domainCoresDict[key].values) == 1}
        return self.assignments

    def find_redundant_cores(self):
        if self.assignments is None:
            self.find_assignments()
        return {key for key in self.binaryCoresDict if
                all([color in self.assignments for color in self.binaryCoresDict[key].colors])}

    def find_variable_cone(self, variables,
                           variableShapes={}):  ## Add variables to domainCoreDict when they are not there!
        for variable in variables:
            if variable + domainCoreSuffix not in self.domainCoresDict:
                self.domainCoresDict[variable + domainCoreSuffix] = engine.get_core()(
                    np.ones(variableShapes[variable]), [variable], variable + domainCoreSuffix)
        variablesQueue = Queue()
        for variable in variables:
            variablesQueue.put(variable)
        coneCores = {}
        while not variablesQueue.empty():
            variable = variablesQueue.get()
            coneCores[variable + domainCoreSuffix] = self.domainCoresDict[variable + domainCoreSuffix]
            if np.sum(self.domainCoresDict[variable + domainCoreSuffix].values) == 1:
                pass
            else:
                for key in self.binaryCoresDict:
                    if variable in self.binaryCoresDict[key].colors:
                        coneCores[key] = self.binaryCoresDict[key]
                        for color in self.binaryCoresDict[key].colors:
                            if color != variable and color + domainCoreSuffix not in coneCores:
                                coneCores[color + domainCoreSuffix] = self.domainCoresDict[color + domainCoreSuffix]
        return coneCores



from tnreason import engine
from tnreason import encoding
from tnreason import algorithms

import pandas as pd

entailedString = "entailed"
contradictingString = "contradicting"
contingentString = "contingent"


class InferenceProvider:
    """
    Answering queries on a distribution by contracting its cores.
    """

    def __init__(self, distribution, contractionMethod=engine.defaultContractionMethod):
        """
        * distribution: Needs to support create_cores() and get_partition_function()
        """
        self.distribution = distribution
        self.contractionMethod = contractionMethod

    def ask_constraint(self, constraint):
        probability = self.ask(constraint, evidenceDict={})
        if probability > 0.9999:
            return entailedString
        elif probability == 0:
            return contradictingString
        else:
            return contingentString

    def ask(self, queryFormula, evidenceDict={}):

        contracted = engine.contract(
            coreDict={
                **self.distribution.create_cores(),
                **encoding.create_evidence_cores(evidenceDict),
                **encoding.create_raw_formula_cores(queryFormula)
            },
            method=self.contractionMethod, openColors=[encoding.get_formula_color(queryFormula)]).values

        return contracted[1] / (contracted[0] + contracted[1])

    def query(self, variableList, evidenceDict={}):
        return engine.contract(method=self.contractionMethod, coreDict={
            **self.distribution.create_cores(),
            **encoding.create_trivial_cores([variable for variable in variableList if
                                             variable not in self.distribution.atoms and variable not in evidenceDict]),
            **encoding.create_evidence_cores(evidenceDict),
        }, openColors=variableList).normalize()

    def exact_map_query(self, variableList, evidenceDict={}):
        distributionCore = self.query(variableList, evidenceDict)
        maxIndex = distributionCore.get_maximal_index()
        return {variable: maxIndex[i] for i, variable in enumerate(distributionCore.colors)}

    def annealed_sample(self, variableList, annealingPattern=[(10, 1)]):
        sampler = algorithms.Gibbs(self.distribution.create_cores())

        sampler.ones_initialization(updateKeys=variableList, shapesDict={variable: 2 for variable in variableList},
                                    colorsDict={variable: [variable] for variable in variableList})

        return sampler.annealed_sample(updateKeys=variableList, annealingPattern=annealingPattern)

    def draw_samples(self, sampleNum, variableList=None, annealingPattern=[(10, 1)], outType="int64"):
        if variableList is None:
            variableList = self.distribution.atoms
        sampleDf = pd.DataFrame(columns=variableList)
        for samplePos in range(sampleNum):
            sampleDf = pd.concat(
                [sampleDf,
                 pd.DataFrame(self.annealed_sample(variableList=variableList, annealingPattern=annealingPattern),
                              index=[samplePos])])
        return sampleDf.astype(outType)



from tnreason import encoding
from tnreason import algorithms


class KnowledgePropagator:
    """
    Evaluates formulas by constraint propagation.
    """
    def __init__(self, knowledgeBase, evidenceDict={}):
        self.atoms = knowledgeBase.atoms
        self.knowledgeCores = {
            **knowledgeBase.create_cores(),
            **encoding.create_evidence_cores(evidenceDict)}

        self.propagator = algorithms.ConstraintPropagator(binaryCoresDict=self.knowledgeCores)

        self.knownHeads = get_evidence_headKeys(evidenceDict) + [
            encoding.get_formula_color(knowledgeBase.facts[key]) + encoding.headCoreSuffix for key in
            knowledgeBase.facts]

    def evaluate(self, variables=None):
        if variables is None:
            variables = self.knownHeads
        self.propagator.initialize_domainCoresDict()
        self.propagator.propagate_cores(coreOrder=variables)
        self.entailedDict = self.propagator.find_assignments()
        return self.entailedDict

    def find_carrying_cores(self, variables=None, variablesShape={}):
        if variables is None:
            variables = self.atoms
        return self.propagator.find_variable_cone(variables, {**variablesShape,
                                                              **{variable: 2 for variable in variables if
                                                                 variable not in variablesShape}})


def get_evidence_headKeys(evidenceDict):
    return [encoding.get_formula_color(key) + encoding.headCoreSuffix for key in evidenceDict if
            evidenceDict[key]] + [
        encoding.get_formula_color(["not", key]) + encoding.headCoreSuffix for key in evidenceDict if
        not evidenceDict[key]
    ]



from tnreason import algorithms
from tnreason import encoding

from tnreason.knowledge import distributions

parameterCoreSuffix = "_parCore"

methodSelectionString = "method"
alsOptionString = "als"
gibbsOptionString = "gibbs"
annealPatternOptionString = "annealingPattern"

headNeuronString = "headNeurons"
architectureString = "architecture"
structureSweepsString = "sweeps"

acceptanceCriterionString = "acceptanceCriterion"

def check_boosting_dict(specDict):
    if methodSelectionString not in specDict:
        raise ValueError("Method not specified for Boosting a formula!")
    if headNeuronString not in specDict:
        raise ValueError("Head Neuron not specified for Boosting a formula!")
    if methodSelectionString not in specDict:
        raise ValueError("Architecture is not specified for Boosting a formula!")

class FormulaBooster:
    """
    Dedicates structure learning to the algorithm subpackage.
    """
    def __init__(self, knowledgeBase, specDict):
        self.knowledgeBase = knowledgeBase
        self.specDict = specDict

    def find_candidate(self, sampleDf):
        """
        Searches for a candidate assignment to the architecture with maximal alignment to the likelihood gradient.
        """
        networkCores = encoding.create_architecture(self.specDict[architectureString], self.specDict[headNeuronString])
        importanceColors = encoding.find_atoms(self.specDict[architectureString])

        empiricalDistribution = distributions.EmpiricalDistribution(sampleDf, importanceColors)

        importanceList = [
            (empiricalDistribution.create_cores(), 1 / empiricalDistribution.get_partition_function(importanceColors)),
            (self.knowledgeBase.create_cores(), -1 / self.knowledgeBase.get_partition_function(importanceColors))]

        colorDims = encoding.find_selection_dimDict(self.specDict[architectureString])
        updateShapes = {key + parameterCoreSuffix: colorDims[key] for key in colorDims}
        updateColors = {key + parameterCoreSuffix: [key] for key in colorDims}
        updateCoreKeys = list(updateShapes.keys())

        ## When alternating least squares used for structure learning
        if self.specDict[methodSelectionString] == alsOptionString:
            sampler = algorithms.ALS(networkCores=networkCores, importanceColors=importanceColors,
                                     importanceList=importanceList, targetCores={})
            sampler.random_initialize(updateKeys=updateCoreKeys, shapesDict=updateShapes, colorsDict=updateColors)
            sampler.alternating_optimization(updateKeys=updateCoreKeys, sweepNum=self.specDict[structureSweepsString])
            solutionDict = sampler.get_color_argmax(updateKeys=updateCoreKeys)
        ## When annealed Gibbs sampling used for structure learning
        elif self.specDict[methodSelectionString] == gibbsOptionString:
            sampler = algorithms.Gibbs(networkCores=networkCores, importanceColors=importanceColors,
                                       importanceList=importanceList,
                                       exponentiated=True)
            sampler.ones_initialization(updateKeys=updateCoreKeys, shapesDict=updateShapes, colorsDict=updateColors)
            if annealPatternOptionString in self.specDict:
                sampleDict = sampler.annealed_sample(updateKeys=updateCoreKeys,
                                                     annealingPattern=self.specDict[annealPatternOptionString])
            elif structureSweepsString in self.specDict:
                sampleDict = sampler.gibbs_sample(updateKeys=updateCoreKeys, sweepNum=self.specDict[structureSweepsString])
            else:
                raise ValueError("Bad parameter specification for Gibbs: {}".format(self.specDict))
            solutionDict = {key[:-len(parameterCoreSuffix)]: int(sampleDict[key]) for key in
                            sampleDict}  # Drop parameterCoreSuffix and ensure int output
        else:
            raise ValueError("Sampling Method {} not known!".format(self.specDict[methodSelectionString]))

        self.candidates = encoding.create_solution_expression(self.specDict[architectureString], solutionDict)

    def test_candidates(self):
        """
        Tests whether to accept the candidate.
        """
        if self.specDict["acceptanceCriterion"] == "always":
            return True
        else:
            raise ValueError("Acceptance Criterion {} not understood.".format(self.specDict[acceptanceCriterionString]))



import numpy as np

def get_unary_tensor(type):
    if type == "id":
        return np.eye(2)
    elif type == "not":
        return create_negation_tensor()


def get_binary_tensor(type):
    if type == "and":
        return create_conjunction_tensor()
    elif type == "or":
        return create_disjunction_tensor()
    elif type == "xor":
        return create_xor_tensor()
    elif type == "imp":
        return create_implication_tensor()
    elif type == "eq":
        return create_biconditional_tensor()
    else:
        raise ValueError("Binary connective {} not understood!".format(type))


def create_truth_vec():
    truthvec = np.zeros(2)
    truthvec[1] = 1
    return truthvec


def create_negation_tensor():
    negation_tensor = np.zeros((2, 2))
    negation_tensor[0, 1] = 1
    negation_tensor[1, 0] = 1
    return negation_tensor


def create_conjunction_tensor():
    and_tensor = np.zeros((2, 2, 2))
    and_tensor[0, 0, 0] = 1
    and_tensor[0, 1, 0] = 1
    and_tensor[1, 0, 0] = 1
    and_tensor[1, 1, 1] = 1
    return and_tensor


def create_disjunction_tensor():
    dis_tensor = np.zeros((2, 2, 2))
    dis_tensor[0, 0, 0] = 1
    dis_tensor[0, 1, 1] = 1
    dis_tensor[1, 0, 1] = 1
    dis_tensor[1, 1, 1] = 1
    return dis_tensor


def create_xor_tensor():
    xor_tensor = np.zeros((2, 2, 2))
    xor_tensor[0, 0, 0] = 1
    xor_tensor[0, 1, 1] = 1
    xor_tensor[1, 0, 1] = 1
    xor_tensor[1, 1, 0] = 1
    return xor_tensor


def create_implication_tensor():
    imp_tensor = np.zeros((2, 2, 2))
    imp_tensor[0, 0, 1] = 1
    imp_tensor[0, 1, 1] = 1
    imp_tensor[1, 0, 0] = 1
    imp_tensor[1, 1, 1] = 1
    return imp_tensor


def create_biconditional_tensor():
    bic_tensor = np.zeros((2, 2, 2))
    bic_tensor[0, 0, 1] = 1
    bic_tensor[0, 1, 0] = 1
    bic_tensor[1, 0, 0] = 1
    bic_tensor[1, 1, 1] = 1
    return bic_tensor













from tnreason.engine.engine_visualization import draw_contractionDiagram

defaultCoreType = "NumpyTensorCore"
defaultContractionMethod = "PgmpyVariableEliminator"

def contract(coreDict, openColors, method=defaultContractionMethod):
    """
    Contractors are initialized with
        * coreDict: Dictionary of colored tensor cores specifying a network
        * openColors: List of colors to leave open in the contraction
    """
    if len(coreDict) == 0:
        return EmptyCore()

    ## Einstein Summation Contractors
    if method == "NumpyEinsum":
        from tnreason.engine.workload_to_numpy import NumpyEinsumContractor
        return NumpyEinsumContractor(coreDict=coreDict, openColors=openColors).contract()
    elif method == "TensorFlowEinsum":
        from tnreason.engine.workload_to_tensorflow import TensorFlowContractor
        return TensorFlowContractor(coreDict=coreDict, openColors=openColors).einsum().to_NumpyTensorCore()
    elif method == "TorchEinsum":
        from tnreason.engine.workload_to_torch import TorchContractor
        return TorchContractor(coreDict=coreDict, openColors=openColors).einsum().to_NumpyTensorCore()
    elif method == "TentrisEinsum":
        from tnreason.engine.workload_to_tentris import TentrisContractor
        return TentrisContractor(coreDict=coreDict, openColors=openColors).einsum().to_NumpyTensorCore()

    ## Variable Elimination Contractors
    elif method == "PgmpyVariableEliminator":
        from tnreason.engine.workload_to_pgmpy import PgmpyVariableEliminator
        return PgmpyVariableEliminator(coreDict=coreDict, openColors=openColors).contract()
    else:
        raise ValueError("Contractor Type {} not known.".format(method))


def get_core(coreType="NumpyTensorCore"):
    if coreType == "NumpyTensorCore":
        from tnreason.engine.workload_to_numpy import NumpyCore
        return NumpyCore
    else:
        raise ValueError("Core Type {} not supported.".format(coreType))

class EmptyCore:
    """
    Output of an empty contraction
    """
    def __init__(self):
        self.values = 1
        self.colors = []
        self.name = "EmptyCore"


from tnreason.encoding.formulas_to_cores import create_formulas_cores, create_raw_formula_cores, get_formula_color, \
    create_head_core, create_evidence_cores, get_atoms, get_all_atoms
from tnreason.encoding.categoricals_to_cores import create_categorical_cores
from tnreason.encoding.neurons_to_cores import create_neuron, create_architecture, find_atoms, find_selection_dimDict, \
    create_solution_expression
from tnreason.encoding.data_to_cores import create_data_cores

from tnreason.encoding.auxiliary_cores import create_trivial_cores, create_trivial_core, create_basis_core, \
    create_random_core

## Core Suffix Nomenclature used in other subpackages
from tnreason.encoding.formulas_to_cores import headCoreSuffix

from tnreason.encoding.storage import save_as_yaml, load_from_yaml



from tnreason import engine

import numpy as np

trivialCoreSuffix = "_trivialCore"


def create_trivial_cores(rawKeys, shapeDict=None, suffix=trivialCoreSuffix):
    """
    Creates dictionary of trivial cores with coordinate 1, which act as neutral placeholders in contractions.
        * rawKeys: List of raw keys (added by suffix)
        * shapeDict: Dictionary of shapes of the trivial core to each rawKey, default: 2
    """
    if shapeDict is None:
        shapeDict = {key: 2 for key in rawKeys}
    return {key + suffix: create_trivial_core(key + suffix, shapeDict[key], [key]) for key in rawKeys}


def create_trivial_core(name, shape, colors):
    return engine.get_core()(np.ones(shape), colors, name)


def create_random_core(name, shape, colors, randomEngine="NumpyUniform"):
    if randomEngine == "NumpyUniform":
        return engine.get_core()(np.random.random(size=shape), colors, name)
    elif randomEngine == "NumpyNormal":
        return engine.get_core()(np.random.normal(size=shape), colors, name)
    else:
        raise ValueError("Random Engine {} not known for core creation!".format(randomEngine))

def create_basis_core(name, shape, colors, numberTuple):
    values = np.zeros(shape=shape)
    values[numberTuple] = 1
    return engine.get_core()(values, colors, name)


import pandas as pd

from tnreason.knowledge import weight_estimation as wees
from tnreason.knowledge import formula_boosting as fb
from tnreason.knowledge import distributions as dist
from tnreason.knowledge import deductive as ded


class HybridLearner:
    """
    Intended to use for extending a Knowledge Base based on data.
    Iterating between:
        - structure learning: Using the FormulaBooster to learn new formulas
        - weight estimation: Using the EntropyMaximizer to adjust the weights to the formulas
    """

    def __init__(self, startKB):
        """
        startKB a knowledge.HybridKnowledgeBase instance representing the current knowledge to be extended.
        """
        self.hybridKB = startKB

    def get_knowledge_base(self):
        return self.hybridKB

    def boost_formula(self, specDict, sampleDf, stepName="_boostStep"):
        """
        Boosting with
        * specDict: Dictionary specification of the Hyperparameters of the Boosting Step:
            - method: Method for structure learning: als or gibbs supported
            - sweeps: Number of sweeps in structure learning
            - architecture: Collection of neurons
            - headNeurons: List of neuronKeys to be used for formula heads
            - calibrationSweeps: Number of sweeps in weight estimation
        * sampleDf: pd.DataFrame storing the data used for the boosting step
        * stepName: Specifies a name suffix for the learned formula to be stored in the HybridKnowledgeBase.
                    Needs to differ for each Step to avoid key conflicts.
        """
        booster = fb.FormulaBooster(self.hybridKB, specDict)
        booster.find_candidate(sampleDf)
        print("Learned formulas: {}".format(booster.candidates))
        if booster.test_candidates():
            print("Accepted formulas.")
            self.hybridKB.include(
                dist.HybridKnowledgeBase(weightedFormulas={
                    candidateKey + stepName: booster.candidates[candidateKey] + [0] for candidateKey in
                    booster.candidates}))
            if "calibrationSweeps" not in specDict:
                specDict["calibrationSweeps"] = 10
            self.calibrate_weights_on_data(specDict, sampleDf=sampleDf)

    def calibrate_weights_on_data(self, specDict, sampleDf, formulaKeys=None):
        if formulaKeys is None:
            formulaKeys = self.hybridKB.weightedFormulas.keys()

        tboFormulas = {
            key: self.hybridKB.weightedFormulas[key][:-1] for key in formulaKeys}

        empDistributionInferer = ded.InferenceProvider(dist.EmpiricalDistribution(sampleDf))
        satDict = {key: empDistributionInferer.ask(tboFormulas[key]) for key in tboFormulas}
        calibrator = wees.EntropyMaximizer(expressionsDict=tboFormulas, satisfactionDict=satDict,
                                           backCores=self.hybridKB.create_hard_cores())

        weightDict, factsDict = calibrator.alternating_optimization(sweepNum=specDict["calibrationSweeps"])
        for key in weightDict:
            self.hybridKB.weightedFormulas[key][-1] = weightDict[key][-1]
        for key in factsDict:
            formula = self.hybridKB.weightedFormulas.pop(key)
            self.hybridKB.facts[key] = formula[:-1]



from setuptools import setup, find_packages

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name="tnreason",
    version="1.0.0",
    author="Alex Goessmann",
    author_email="alex.goessmann@web.de",
    description="A package for reasoning based on encoding networks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence"
    ],
    install_requires=[
        "numpy>=1.17",  # was 1.23.4",
        "pandas>=1.0.0",
        "networkx",
        "pyyaml",
        "rdflib",
        "matplotlib",
        "pgmpy",
        "importlib-resources"
    ],
    python_requires=">=3",
    license="AGPL-3.0",
    url="https://github.com/EnexaProject/enexa-tensor-reasoning",
    keywords="markov logic networks, tensor networks, alternating least squares, gibbs sampling, inductive logic programming"
)



## Moment Matching and Binary Propagation not used in knowledge so far!

from tnreason.algorithms.alternating_least_squares import ALS
from tnreason.algorithms.gibbs_sampling import Gibbs
from tnreason.algorithms.constraint_propagation import ConstraintPropagator


from tnreason import encoding
from tnreason import engine

probFormulasKey = "weightedFormulas"
logFormulasKey = "facts"
categoricalsKey = "categoricalConstraints"
evidenceKey = "evidence"

"""
Distributions are Markov Networks with two methods:
    * create_cores(): returning the factor cores
    * get_partition_function(allAtoms): returning the partition function given the atomic variables of interest
"""

class EmpiricalDistribution:
    """
    Inferable (by HybridInferer) empirical distributions
    """

    def __init__(self, sampleDf, atomKeys=None):
        """
        * sampleDf: pd.DataFrame containing the samples defining the empirical distributions
        * atomKeys: List of columns of sampleDf to be recognized as atoms
        """
        if atomKeys is None:
            atomKeys = list(sampleDf.columns)
        self.atoms = atomKeys
        self.sampleDf = sampleDf
        self.dataNum = sampleDf.values.shape[0]

    def __str__(self):
        return "Empirical Distribution with {} samples of atoms {}.".format(self.dataNum, self.atoms)

    def create_cores(self):
        return encoding.create_data_cores(self.sampleDf, self.atoms)

    def get_partition_function(self, allAtoms=[]):
        unseenAtomNum = len([atom for atom in allAtoms if atom not in self.atoms])
        return (self.dataNum * (2 ** unseenAtomNum))


class HybridKnowledgeBase:
    """
    Inferable (by HybridInferer) Knowledge Base. Generalizes Markov Logic Network by further dedicated cores
    """

    def __init__(self, weightedFormulas={}, facts={}, categoricalConstraints={}, evidence={}):
        self.weightedFormulas = weightedFormulas
        self.facts = facts
        self.categoricalConstraints = categoricalConstraints
        self.evidence = evidence

        self.find_atoms()

    def __str__(self):
        return "\n".join([
            "Hybrid Knowledge Base consistent of",
            "######## probabilistic formulas:",
            *[encoding.get_formula_color(expression[:-1]) + " with weight " + str(expression[-1]) for expression in
              self.weightedFormulas.values()],
            "######## logical formulas:",
            *[encoding.get_formula_color(expression) for expression in self.facts.values()],
            "######## categorical variables:",
            *[key + " selecting one of " + " ".join(self.categoricalConstraints[key]) for key in
              self.categoricalConstraints]
        ])

    def find_atoms(self):
        """
        Identifies the atoms of the Knowledge Base
        """
        self.atoms = encoding.get_all_atoms(
            {**{key: self.weightedFormulas[key][:-1] for key in self.weightedFormulas},
             **self.facts})
        for constraintKey in self.categoricalConstraints:
            for atom in self.categoricalConstraints[constraintKey]:
                if atom not in self.atoms:
                    self.atoms.append(atom)
        for eKey in self.evidence:
            if eKey not in self.atoms:
                self.atoms.append(eKey)
        self.atoms = list(self.atoms)

    def from_yaml(self, loadPath):
        modelSpec = encoding.load_from_yaml(loadPath)
        if probFormulasKey in modelSpec:
            self.weightedFormulas = modelSpec[probFormulasKey]
        if logFormulasKey in modelSpec:
            self.facts = modelSpec[logFormulasKey]
        if categoricalsKey in modelSpec:
            self.categoricalConstraints = modelSpec[categoricalsKey]
        if evidenceKey in modelSpec:
            self.evidence = modelSpec[evidenceKey]
        self.find_atoms()

    def to_yaml(self, savePath):
        encoding.storage.save_as_yaml({
            probFormulasKey: self.weightedFormulas,
            logFormulasKey: self.facts,
            categoricalsKey: self.categoricalConstraints,
            evidenceKey: self.evidence
        }, savePath)

    def include(self, secondHybridKB):
        self.weightedFormulas = {**self.weightedFormulas,
                                 **secondHybridKB.weightedFormulas}
        self.facts = {**self.facts,
                      **secondHybridKB.facts}
        self.categoricalConstraints = {**self.categoricalConstraints,
                                       **secondHybridKB.categoricalConstraints}
        self.evidence = {**self.evidence,
                         **secondHybridKB.evidence}
        self.find_atoms()

    def create_cores(self):
        return {**encoding.create_formulas_cores({**self.weightedFormulas, **self.facts}),
                **encoding.create_evidence_cores(self.evidence),
                **encoding.create_categorical_cores(self.categoricalConstraints)}

    def get_partition_function(self, allAtoms=[]):
        unseenAtomNum = len([atom for atom in allAtoms if atom not in self.atoms])
        return (engine.contract(coreDict=self.create_cores(), openColors=[]).values
                * (2 ** unseenAtomNum))

    def create_hard_cores(self):
        """
        Returns the cores posing hard logical constraints on the worlds to be models
        """
        return {**encoding.create_formulas_cores(self.facts),
                **encoding.create_evidence_cores(self.evidence),
                **encoding.create_categorical_cores(self.categoricalConstraints)}

    def is_satisfiable(self):
        """
        Decides whether the Knowledge Base is satisfiable, i.e. whether a model exists
        """
        return engine.contract(coreDict=self.create_hard_cores(),
                               openColors=[]).values > 0
